/*
  This program demonstrates moving the camera around.
  It takes three "pictures" of a scene, a front view,
  a right side view, and a top view of the scene.

  The three pictures are all stored in the same
  framebuffer. This demonstrates the idea of
  "viewports" within a framebuffer.
*/
import java.awt.Color;

public class MainViews
{
   public static void main(String[] args)
   {
      // Create several Model objects.
      Model m1 = new TetrahedronModel();
      for (Triangle t : m1.triangleList)
      {
         t.v[0].setColorRandom();
         t.setColor( t.v[0].getColor() );
      }
      m1.modelTranslate(-2, 0, 1);
      m1.modelScale(2, 2, 2);


      Model m2 = new OctahedronModel();
      for (Triangle t : m2.triangleList)
      {
         t.v[0].setColorRandom();
         t.setColor( t.v[0].getColor() );
      }
      m2.modelTranslate(2, 2, -1);
      m2.modelScale(2, 2, 2);


      Model m3 = new ConeModel(4, 4, 1, 5);
      for (Triangle t : m3.triangleList)
      {
         t.v[0].setColorRandom();
         t.setColor( t.v[0].getColor() );
      }


      Model floor = new PanelModel(-5, 5, -5, 5, Color.white, Color.black);
      floor.doLighting = false;

      // Create the Scene object that we shall render
      Scene scene = new Scene();

      // Add the models to the Scene.
      scene.modelList.add(m1);
      scene.modelList.add(m2);
      scene.modelList.add(m3);
      scene.modelList.add(floor);

      // Give the Scene a view matrix.
      scene.view2Identity();

      // Give the scene a projection matrix.
      scene.proj2Identity();

      // Set up the scene's view frustum.
    //double fov    = 60.0;
    //double aspect = 1.0;
      double near   = 1.0;
      double far    = 20.0;
    //scene.projPerspective(fov, aspect, near, far);
      double left    = -3;
      double right   =  3;
      double bottom  = -3;
      double top     =  3;
      scene.projOrtho(left, right, bottom, top, near, far);


      // Create a framebuffer to render our scene into.
      int vp_width  = 512;
      int vp_height = 512;
      FrameBuffer fb = new FrameBuffer(vp_width, vp_height);

      for (int i = 0; i <= 72; i++)
      {
         // Give the framebuffer a nice background color.
         fb.clearFB(java.awt.Color.darkGray);

         // Set a viewport within the frambuffer.
         fb.setViewport(0, 0, 256, 256); //upper-left-hand-corner, width, height
         // Give the viewport a nice background color.
         fb.clearVP(java.awt.Color.darkGray.brighter().brighter());
         // Set up the scene's "camera" (front view).
         scene.view2Identity();
         double eyex=0,    eyey=3,    eyez=4;       // where the camera is
         double centerx=0, centery=3, centerz=0;    // where the camera is pointed
         double upx=0,     upy=1,     upz=0;        // where the "top" of the camera is pointed
         scene.viewLookAt(eyex,    eyey,    eyez,
                          centerx, centery, centerz,
                          upx,     upy,     upz);
         // Render our scene into the view port within the frame buffer.
         SceneRender.renderScene(scene, fb);

         // Set a new viewport within the frambuffer.
         fb.setViewport(256, 0, 256, 256);
         // Give the viewport a nice background color.
         fb.clearVP(java.awt.Color.darkGray.brighter().brighter());
         // Change the view (side view).
         scene.view2Identity();
         eyex=4;    eyey=3;    eyez=0;
         centerx=0; centery=3; centerz=0;
         upx=0;     upy=1;     upz=0;
         scene.viewLookAt(eyex,    eyey,    eyez,
                          centerx, centery, centerz,
                          upx,     upy,     upz);
         // Render again, but into the new viewport.
         SceneRender.renderScene(scene, fb);

         // Set a new viewport within the frambuffer.
         fb.setViewport(0, 255, 256, 256);
         // Give the viewport a nice background color.
         fb.clearVP(java.awt.Color.darkGray.brighter().brighter());
         // Change the view (top view).
         scene.view2Identity();
         eyex=0;    eyey=6;    eyez=0;
         centerx=0; centery=0; centerz=0;
         upx=0;     upy=0;     upz=-1;
         scene.viewLookAt(eyex,    eyey,    eyez,
                          centerx, centery, centerz,
                          upx,     upy,     upz);
         // Render again, but into the new viewport.
         SceneRender.renderScene(scene, fb);

         // Save the resulting image in a file.
         fb.dumpFB2File(String.format("PPM_ThreeViews_Frame%02d.ppm", i));

         // rotate the models
         m1.modelRotate(5, 0,1,0);
         m2.modelRotate(5, 1,0,0);
         m3.modelRotate(5, 0,1,0);
      }
   }
}