/**
   A Model data structure is mainly a list of Triangle objects
   plus a model matrix (which determines the models location in
   world coordinates).

   The Model also contains any textures used by some of its triangles.
*/
import java.util.LinkedList;

public class Model
{
   public LinkedList<Triangle> triangleList = new LinkedList<Triangle>();
   public Texture[] textures = new Texture[256];
   public int textureCount = 0;
   public boolean bilerp_flag;  // determines which rendering mode to use for textures.
   public Matrix modelMatrix;
   public boolean doBackFaceCulling;
   public boolean frontFacingIsCCW;
   public boolean doLighting;
   public boolean doSmoothLighting;
   public int  doOpenglFlatLighting;  // flat lighting in the OpenGL style
   public boolean useHalfVector;
   public boolean doSmoothShading;
   public boolean doTransparancy;

   public Matrix viewMatrix;  // These are not initialized until
   public Matrix projMatrix;  // the model gets put into a Scene.
   public Light light;


   public Model()
   {
      bilerp_flag = false;
      modelMatrix = new Matrix(1.0); // identity matrix
      doBackFaceCulling = false;
      frontFacingIsCCW  = true;
      doLighting        = false;
      doSmoothLighting  = true;
      doOpenglFlatLighting = 0;
      useHalfVector     = false;
      doSmoothShading   = true;
      doTransparancy    = false;
   }// default constructor


   /**
      Duplicate a Model object.
   */
   public Model(Model model1)
   {
      // Copy information from model1 to model.
      for (Triangle t : model1.triangleList)
      {
         this.triangleList.add( new Triangle(t) );
      }
      for (int i = 0; i < model1.textureCount; i++)
      {
         this.textures[i] = model1.textures[i];  // this is NOT a deep copy
      }
      this.textureCount         = model1.textureCount;
      this.bilerp_flag          = model1.bilerp_flag;
      this.modelMatrix          = model1.modelMatrix; // NOT a deep copy
      this.viewMatrix           = model1.viewMatrix;  // NOT a deep copy
      this.projMatrix           = model1.projMatrix;  // NOT a deep copy
      this.light                = model1.light;       // NOT a deep copy
      this.frontFacingIsCCW     = model1.frontFacingIsCCW;
      this.doBackFaceCulling    = model1.doBackFaceCulling;
      this.doLighting           = model1.doLighting;
      this.doSmoothLighting     = model1.doSmoothLighting;
      this.doOpenglFlatLighting = model1.doOpenglFlatLighting;
      this.useHalfVector        = model1.useHalfVector;
      this.doSmoothShading      = model1.doSmoothShading;
      this.doTransparancy       = model1.doTransparancy;
   }//duplicating constructor


   public void addTexture(Texture tex)
   {
      textures[textureCount++] = tex;
   }


   public void addViewMatrix(Matrix viewMatrix) // Called by the renderer just before
   {                                            // this model is rendered.
      this.viewMatrix = viewMatrix;
   }


   public void addProjMatrix(Matrix projMatrix) // Called by the renderer just before
   {                                            // this model is rendered.
      this.projMatrix = projMatrix;
   }


   public void addLight(Light light) // Called by the renderer just before
   {                                 // this model is rendered.
      this.light = light;
   }


   public void model2Identity()
   {
      modelMatrix = new Matrix(1.0);
   }


   public void modelMult(Matrix m)
   {
      modelMatrix = modelMatrix.times(m);
   }


   public void modelRotate(double theta, double x, double y, double z)
   {
      modelMatrix = modelMatrix.times(new Matrix(theta, x, y, z));
   }


   public void modelScale(double x, double y, double z)
   {
      modelMatrix = modelMatrix.times(new Matrix(x, y, z));
   }


   public void modelTranslate(double x, double y, double z)
   {
      modelMatrix = modelMatrix.times(new Matrix(new Vector(x, y, z)));
   }


   /**
      For debugging.
   */
   public String toString(int fields)
   {
      String result = toString();
      //result = "Printing out this Model's " + triangleList.size() + " Triangles:\n";
      for (Triangle t : this.triangleList)
      {
         result += t.toString(fields);
      }
      //result += "Done printing out Model\n";
      return result;
   }


   public String toString()
   {
      String result = "";
      result += "This Model has " + triangleList.size() + " triangles\n";;
      result += "This Model has " + textureCount        + " textures.\n";
      result += "doBackFaceCulling = "    + doBackFaceCulling + "\n";
      result += "frontFacingIsCCW = "     + frontFacingIsCCW + "\n";
      result += "doLighting = "           + doLighting + "\n";
      result += "doSmoothLighting = "     + doSmoothLighting + "\n";
      result += "doOpenglFlatLighting = " + doOpenglFlatLighting + "\n";
      result += "useHalfVector = "        + useHalfVector + "\n";
      result += "doSmoothShading = "      + doSmoothShading + "\n";
      result += "doTransparancy = "       + doTransparancy + "\n";
      return result;
   }
}