/**
   A Scene data structure is mainly a list of
   Model data structures plus a Light data structure
   and a view matrix (which determines the scene's
   "camera" location) and a projection matrix (which
   determines the camera's view volume).
*/
import java.util.LinkedList;

public class Scene
{
   public LinkedList<Model> modelList = new LinkedList<Model>();
   public Matrix projMatrix;
   public Matrix viewMatrix;
   public Light light;          // data structure that describes a light

   public Scene()
   {
      projMatrix = new Matrix(1.0); // identity matrix
      viewMatrix = new Matrix(1.0);
      light = null;
   }// default constructor


   /**
      Duplicate a Scene object.
   */
   public Scene(Scene scene1)
   {
      // Copy information from scene1 to scene.
      for (Model m : scene1.modelList)
      {
         this.modelList.add( new Model(m) );
      }
      this.projMatrix = scene1.projMatrix; // NOT a deep copy
      this.viewMatrix = scene1.viewMatrix; // NOT a deep copy
      this.light = scene1.light;           // NOT a deep copy
   }//duplicating constructor


   public void addModel(Model model)
   {
      modelList.add(model);
   }


   public void proj2Identity()
   {
      projMatrix = new Matrix(1.0);
   }


   public void projMult(Matrix m)
   {
      projMatrix = projMatrix.times(m);
   }


   public void projPerspective(double fovy, double aspect, double zNear, double zFar)
   {
      //  http://www.opengl.org/sdk/docs/man2/xhtml/gluPerspective.xml
      double f = 1.0/Math.tan((Math.PI/180.0)*fovy/2.0);
      projMatrix = projMatrix.times(
                        new Matrix(new Vector(f/aspect, 0.0, 0.0, 0.0),
                        new Vector(0.0, f, 0.0, 0.0),
                        new Vector(0.0, 0.0, (zFar+zNear)/(zNear-zFar), -1.0),
                        new Vector(0.0, 0.0, (2.0*zFar*zNear)/(zNear-zFar), 0.0)));
   }


   public void projFrustum(double left, double right, double bottom, double top, double near, double far)
   {
      //  http://www.opengl.org/sdk/docs/man2/xhtml/glFrustum.xml
      double A = (right+left)/(right-left);
      double B = (top+bottom)/(top-bottom);
      double C =   (far+near)/(far-near);
      double D = (2*far*near)/(far-near);
      projMatrix = projMatrix.times(new Matrix(
                        new Vector((2.0*near)/(right-left), 0.0, 0.0, 0.0),
                        new Vector(0.0, (2.0*near)/(top-bottom), 0.0, 0.0),
                        new Vector(  A,   B,  C, -1.0),
                        new Vector(0.0, 0.0,  D,  0.0)));
   }


   public void projOrtho(double left, double right, double bottom, double top, double near, double far)
   {
      //  http://www.opengl.org/sdk/docs/man2/xhtml/glOrtho.xml
      double tx = -(right+left)/(right-left);
      double ty = -(top+bottom)/(top-bottom);
      double tz =   -(far+near)/(far-near);
      projMatrix = projMatrix.times(
                        new Matrix(new Vector((2.0)/(right-left), 0.0, 0.0, 0.0),
                        new Vector(0.0, (2.0)/(top-bottom), 0.0, 0.0),
                        new Vector(0.0, 0.0, (-2.0)/(far-near), 0.0),
                        new Vector(tx, ty, tz, 1.0)));
   }


   public void view2Identity()
   {
      viewMatrix = new Matrix(1.0);
   }


   public void viewLookAt(double eyex,    double eyey,    double eyez,
                          double centerx, double centery, double centerz,
                          double upx,     double upy,     double upz)
   {
      //  http://www.opengl.org/sdk/docs/man2/xhtml/gluLookAt.xml
      Vector F  = new Vector(centerx - eyex, centery - eyey, centerz - eyez);
      Vector UP = new Vector(upx, upy, upz);
      Vector f  = F.normalize();
      Vector up = UP.normalize();
      Vector s  = f.crossProduct(up);
      Vector u  = s.crossProduct(f);
      viewMatrix = viewMatrix.times(new Matrix(new Vector(s.x, u.x, -f.x, 0.0),
                                               new Vector(s.y, u.y, -f.y, 0.0),
                                               new Vector(s.z, u.z, -f.z, 0.0),
                                               new Vector(0.0, 0.0,  0.0, 1.0)));

      viewMatrix = viewMatrix.times(new Matrix(new Vector(-eyex, -eyey, -eyez)));
   }


   public void viewRotate(double theta, double x, double y, double z)
   {
      viewMatrix = viewMatrix.times(new Matrix(theta, x, y, z));
   }


   public void viewTranslate(double x, double y, double z)
   {
      viewMatrix = viewMatrix.times(new Matrix( new Vector(x, y, z)));
   }


   public String toString()
   {
      String result = "";
      result += "This Scene has " + modelList.size() + " models\n";;
      result += "light = "                + light + "\n";
      int i = 0;
      for (Model m : modelList)
      {
         result += "Model " + (i++) + "\n";
         result += m.toString();
      }
      return result;
   }
}