/**

*/
import java.util.Random;
import java.awt.Color;

public class Vertex
{
   public double x, y, z, w;  // a vertex in homogenous coordinates
   public double r, g, b, a;  // color at this vertex (including transparency)
   public double s, t;        // texture coordinates at this vertex
   public Vector n;           // normal vector at this vertex


   public Vertex()
   {
      this(0.0, 0.0, 0.0, 1.0);
   }


   public Vertex(double x, double y, double z)
   {
      this(x, y, z, 1.0);
   }


   public Vertex(float[] v)
   {
      this(v[0], v[1], v[2], v[3]);
   }


   public Vertex(double x, double y, double z, double w)
   {
      set(x, y, z, w);

      setColor(0.0, 0.0, 0.0, 0.0);

      this.s = 0.0;
      this.t = 0.0;

      this.n = new Vector();
   }


   public Vertex(Vertex v) // a "copy constructor"
   {
      set(v.x, v.y, v.z, v.w);

      setColor(v.r, v.g, v.b, v.a);

      this.s = v.s;
      this.t = v.t;

      this.n = new Vector(v.n);
   }


   public Vector minus(Vertex v) // vertex minus vertex returns a vector
   {
      return new Vector( x-(v.x), y-(v.y), z-(v.z), w-(v.w) );
   }


   private void set(double x, double y, double z)
   {
      set(x, y, z, 1.0);
   }


   private void set(double x, double y, double z, double w)
   {
      this.x = x;
      this.y = y;
      this.z = z;
      this.w = w;
   }


   /**
      Set the color and transparency of this vertex to (r,g,b,a).
   */
   public void setColor(double r, double g, double b, double a)
   {
      this.r = r;
      this.g = g;
      this.b = b;
      this.a = a;

      if ( (r < 0.0) || (r > 1.0)
        || (g < 0.0) || (g > 1.0)
        || (b < 0.0) || (b > 1.0)
        || (a < 0.0) || (a > 1.0) )
      {
         System.err.println("ERROR! Invalid double color for vertex");
         System.err.println( this.toString(3) );
         System.exit(-1);
      }
   }


   public void setColor(double r, double g, double b)
   {
      setColor(r, g, b, 0.0);
   }


   public void setColor(Vertex v) // copy the color from v to this vertex
   {
      double r = v.r;
      double g = v.g;
      double b = v.b;
      setColor(r, g, b, 0.0);
   }


   public void setColorRandom()
   {
      Random generator = new Random();
      double r = generator.nextDouble();
      double g = generator.nextDouble();
      double b = generator.nextDouble();
      setColor(r, g, b, 0.0);
   }


   public void setColor(int r, int g, int b, int a)
   {
      this.r = ((float)r)/(float)255;
      this.g = ((float)g)/(float)255;
      this.b = ((float)b)/(float)255;
      this.a = ((float)a)/(float)255;

      if ( (r < 0) || (r > 255)
        || (g < 0) || (g > 255)
        || (b < 0) || (b > 255)
        || (a < 0) || (a > 255))
      {
         System.err.println("ERROR! Invalid int color for vertex");
         System.err.println( this.toString(3) );
         System.exit(-1);
      }
   }


   public void setColor(int r, int g, int b)
   {
      setColor(r, g, b, 0);
   }


   public void setColor(Color c)
   {
      setColor(c.getRed(), c.getGreen(), c.getBlue(), 0);
   }


   public Color getColor()
   {
      return new Color((float)this.r, (float)this.g, (float)this.b);
   }


   /**
      Set the texture coordinate of this vertex to (s, t).
   */
   public void setCoords(double s, double t)
   {
      this.s = s;
      this.t = t;
   }


   public void setNormal(Vector n)
   {
      this.n = n;
   }


   /**
      Scalar times vertex.
      Needed for the perspective division step in the rendering pipeline.
   */
   public Vertex times(double scalar) // scalar times Vertex
   {
      Vertex temp = new Vertex(scalar*x, scalar*y, scalar*z, scalar*w);
      temp.r = r;
      temp.g = g;
      temp.b = b;
      temp.a = a;
      temp.s = s;
      temp.t = t;
      temp.n = n;
      return temp;
   }


   /**
      For debugging.
   */
   public String toString(int fields)
   {
      String result = "";
      int p = 5;     // the precision for the following format strings
      int t = p + 4; // the width for the folowing format strings
      if (1 == (fields & 1))
      {
         // Here is one way to get programable precision and width.
         String format = "(x,y,z,w) = (% "+t+"."+p+"f  % "+t+"."+p+"f  % "+t+"."+p+"f  % "+t+"."+p+"f)\n";
         result += String.format( format, x, y, z, w);
       //result += String.format("(x,y,z,w) = (% .5f  % .5f  % .5f  % .5f)\n", x, y, z, w);
      }
      if (2 == (fields & 2))
      {
         result += String.format("<r,g,b,a> = <% .5f  % .5f  % .5f  % .5f>\n", r, g, b, a);
      }
      if (4 == (fields & 4))
      {
         result += String.format("    {s,t} = {% .5f  % .5f}\n", s, t);
      }
      if (8 == (fields & 8))
      {
         result += String.format("n -> \n");
         result += n.toString();
      }
      return result;
   }


   public String toString()
   {
      // Here is one way to get programable precision and width.
      int p = 5;     // the precision for the following format string
      int t = p + 4; // the width for the folowing format string
      String format =      "(x,y,z,w) = (% "+t+"."+p+"f  % "+t+"."+p+"f  % "+t+"."+p+"f  % "+t+"."+p+"f)\n";
      return String.format( format, x, y, z, w);
    //return String.format("(x,y,z,w) = (% .5f  % .5f  % .5f  % .5f)", x, y, z, w);
   }
}