/**
   Cull from the list of triangles any triangle that is "back facing".

   For a good explanation of orientation, forward and backward facing polygons,
   and back face culling, see Chapter 9 of
      "Computer Graphics Through OpenGL: From Theory to Experiments "
       by Sumanta Guha, Chapman and Hall/CRC, 2010.
  That chapter is available at
       http://www.sumantaguha.com/files/materials/chapter9.pdf
*/

public class P3_BackFaceCulling
{
   /**
      Walk the list of Triangle objects.
      Determine if a triangle is forward or
      backward facing. Delete any triangle
      that is backward facing.
   */
   public static void P3_BackFaceCulling(Model model)
   {
      if (model.doBackFaceCulling)
      {
         java.util.Iterator<Triangle> it = model.triangleList.iterator();

         while ( it.hasNext() )
         {
            Triangle t = it.next();
            if ( ! triangleIsFrontFacing(model.frontFacingIsCCW, t) )
            {
               it.remove(); // unlink this triangle
            }
         }
      }
   }


   public static boolean triangleIsFrontFacing(boolean frontFacingIsCCW, Triangle t)
   {
      boolean result = triangleIsCCW(t);
      if ( ! frontFacingIsCCW )
         result = ! result;
      return result;
   }


   public static boolean triangleIsCCW(Triangle t)
   {
      // We only need to know the z component of the cross product
      // of the two "edge vectors", v1-v0 and v2-v0. This works
      // because we are in view coordinates, with the camera at
      // the origin looking down the negative z-axis.
      double x0 = t.v[0].x;
      double x1 = t.v[1].x;
      double x2 = t.v[2].x;
      double y0 = t.v[0].y;
      double y1 = t.v[1].y;
      double y2 = t.v[2].y;
      double zValueOfCrossProduct = (x1-x0)*(y2-y0) - (y1-y0)*(x2-x0);

      //System.err.printf("zValueOfCrossProduct = % .6f\n", zValueOfCrossProduct);

      return (zValueOfCrossProduct >= 0);
   }


   /*
   public static boolean triangleIsCCW(Triangle *t)
   {
      // Here is another way to calculate the exact same thing.
      // Compute the signed area of the triangle projected into the xy-plane.
      double x0 = t.v[0].x;
      double x1 = t.v[1].x;
      double x2 = t.v[2].x;
      double y0 = t.v[0].y;
      double y1 = t.v[1].y;
      double y2 = t.v[2].y;
      double determinant = (x1-x0)*(y2-y0) - (y1-y0)*(x2-x0);

      //System.err.printf("determinant = % .6f\n", determinant);

      return (determinant > 0);
   }
   */


   /*
   public static boolean triangleIsCCW(Triangle *t)
   {
      double x0 = t.v[0].x;
      double x1 = t.v[1].x;
      double x2 = t.v[2].x;
      double y0 = t.v[0].y;
      double y1 = t.v[1].y;
      double y2 = t.v[2].y;
      double z0 = t.v[0].z;
      double z1 = t.v[1].z;
      double z2 = t.v[2].z;
      double determinant = x0*(y1*z2-y2*z1)-x1*(y0*z2-y2*z0)+x2*(y0*z1-y1*z0);

      //System.err.printf("determinant = % .6f\n", determinant);

      return (determinant > 0);
   }
   */
}