/**
   Compute the shade (i.e., color) of each vertex
   as determined by all of the lighting data.
*/

public class P4_LightingSimple_ver1
{
   public static void P4_Lighting(Model model)
   {
      if ( model.doLighting && (null != model.light) )
      {
         doFlatLighting(model);
      }
   }


   /**
      Walk the list of Triangle objects.
      For each Triangle object, use the Scene's light
      data and the Triangle's material properties to
      compute color information for the triangle.
      This version computes a normal for each traingle
      and it does a lighting calculation once, at the
      center of the triangle. The result of that lighting
      calculation is copied to each of the three vertices.
   */
   public static void doFlatLighting(Model model)
   {
      for (Triangle t : model.triangleList)
      {
         // Compute a normal vector, N, for this triangle. We need to be careful
         // that the normal vector has the right "orientation". We want the normal
         // vector for a CCW triangle (which is the default for a "forward facing
         // triangle") to point towards the origin (in view coordinates). Compare
         // with P3_BackFaceCulling.cpp. Another way to put this is that for a
         // closed surface made up of CCW triangles, the normal vector should
         // point outside of the surface.
         Vector p1 = new Vector(t.v[1].x - t.v[0].x,
                                t.v[1].y - t.v[0].y,
                                t.v[1].z - t.v[0].z);
         Vector p2 = new Vector(t.v[2].x - t.v[0].x,
                                t.v[2].y - t.v[0].y,
                                t.v[2].z - t.v[0].z);
         Vector N = p1.crossProduct(p2);
         N = N.normalize();

         Vector p0 = new Vector(t.v[0].x, t.v[0].y, t.v[0].z);

         // Compute the location of the center of this triangle.
         // p = p0 + (v1-v0)/3 + (v2-v0)/3 is our center (position) vector for this triangle.
         Vector p = p0.plus( p1.times(1.0/3.0).plus(p2.times(1.0/3.0)) );

         // We need the unit light vector, L, from our center position to the light source.
         // Here is the light source as a vertex in world coordinates.
         Vertex S = new Vertex(model.light.pos);
         // Convert light source from world coordinates to view coordinates.
         Matrix m = model.viewMatrix;
         S = m.times(S);
         // Convert the vertex to a vector.
         Vector S2 = new Vector(S.x, S.y, S.z);
         // Compute L = S2 - p, the vector from our center position to the light source.
         Vector L = S2.plus( p.times(-1.0) );
         // normalize L
         L = L.normalize();

         // We need the reflection vector R = 2(L dot N)*N - L.
         Vector R = N.times( 2*L.dotProduct(N) ).plus( L.times(-1.0) );

         // We need the unit view vector, V, that points from the center position
         // to the origin (in view coordinates).
         Vector V = p.normalize().times(-1.0);

         // Compute the color at the center position using normal, light, reflection, and view vectors,
         //    c = k_a * c_a + sum_{all lights}( k_d*c_d*(L dot N) + k_s*c_s*(R dot V)^s ) )

         // pre compute some numbers
         double s = t.mtrl.s; //shininess coefficient
         double diffuseAngle = L.dotProduct(N);
         diffuseAngle = (diffuseAngle < 0) ? 0.0 : diffuseAngle;
         double shininess = R.dotProduct(V);
         shininess = (shininess < 0) ? 0.0 : Math.pow(shininess, s);

         // do red
         double kar = t.mtrl.ambient[0];
         double kdr = t.mtrl.diffuse[0];
         double ksr = t.mtrl.specular[0];
         double car = model.light.ambient[0];
         double cdr = model.light.diffuse[0];
         double csr = model.light.specular[0];
         double r = kar*car + kdr*cdr*diffuseAngle + ksr*csr*shininess;
         r = (r <= 1.0) ? r : 1.0;  // clamp the color value at 1.0
         // copy the result to all three vertices
         t.v[0].r = r;
         t.v[1].r = r;
         t.v[2].r = r;


         // do green
         double kag = t.mtrl.ambient[1];
         double kdg = t.mtrl.diffuse[1];
         double ksg = t.mtrl.specular[1];
         double cag = model.light.ambient[1];
         double cdg = model.light.diffuse[1];
         double csg = model.light.specular[1];
         double g = kag*cag + kdg*cdg*diffuseAngle + ksg*csg*shininess;
         g = (g <= 1.0) ? g : 1.0;  // clamp the color value at 1.0
         // copy the result to all three vertices
         t.v[0].g = g;
         t.v[1].g = g;
         t.v[2].g = g;

         // do blue
         double kab = t.mtrl.ambient[2];
         double kdb = t.mtrl.diffuse[2];
         double ksb = t.mtrl.specular[2];
         double cab = model.light.ambient[2];
         double cdb = model.light.diffuse[2];
         double csb = model.light.specular[2];
         double b = kab*cab + kdb*cdb*diffuseAngle + ksb*csb*shininess;
         b = (b <= 1.0) ? b : 1.0;  // clamp the color value at 1.0
         // copy the result to all three vertices
         t.v[0].b = b;
         t.v[1].b = b;
         t.v[2].b = b;
      }
   }
}