/**
   Compute the shade (i.e., color) of each vertex
   as determined by all of the lighting data.
*/

public class P4_LightingSimple_ver2
{
   public static void P4_Lighting(Model model)
   {
      if ( model.doLighting && (null != model.light) )
      {
         doSmoothLighting(model);
      }
   }


   /**
      Walk the list of Triangle objects.
      For each Vertex in each Triangle object, use
      the Scene's light data, the Triangle's material
      properties, and the Vertex's normal vector to
      compute color information for that vertex.
   */
   public static void doSmoothLighting(Model model)
   {
      // Compute the transformation matrix that transforms normal
      // vectors from model coordinates into view coordinates.
      // Get the needed 3-by-3 part of the model matrix.
      double m11 = model.modelMatrix.v1.x, m12 = model.modelMatrix.v2.x, m13 = model.modelMatrix.v3.x,
             m21 = model.modelMatrix.v1.y, m22 = model.modelMatrix.v2.y, m23 = model.modelMatrix.v3.y,
             m31 = model.modelMatrix.v1.z, m32 = model.modelMatrix.v2.z, m33 = model.modelMatrix.v3.z;
      // Get the needed 3-by-3 part of the view matrix.
      double v11 = model.viewMatrix.v1.x, v12 = model.viewMatrix.v2.x, v13 = model.viewMatrix.v3.x,
             v21 = model.viewMatrix.v1.y, v22 = model.viewMatrix.v2.y, v23 = model.viewMatrix.v3.y,
             v31 = model.viewMatrix.v1.z, v32 = model.viewMatrix.v2.z, v33 = model.viewMatrix.v3.z;
      // Compute V*M, the 3-by-3 modelview matrix.
      double mv11 = v11*m11 + v12*m21 + v13*m31,
             mv12 = v11*m12 + v12*m22 + v13*m32,
             mv13 = v11*m13 + v12*m23 + v13*m33,
             mv21 = v21*m11 + v22*m21 + v23*m31,
             mv22 = v21*m12 + v22*m22 + v23*m32,
             mv23 = v21*m13 + v22*m23 + v23*m33,
             mv31 = v31*m11 + v32*m21 + v33*m31,
             mv32 = v31*m12 + v32*m22 + v33*m32,
             mv33 = v31*m13 + v32*m23 + v33*m33;
      // Compute (V*M)^{-T}, the inverse transpose of the 3-by-3 modelview matrix.
      // We are using the trick here of computing the inverse using cofactors,
      // but without dividing by the determinant. So this is not the true inverse,
      // but we only need our "inverse" to point our vectors in the correct direction.
      // We will then normalize our vectors to length 1.
      Vector col1 = new Vector(mv22*mv33-mv23*mv32, mv13*mv32-mv12*mv33, mv12*mv23-mv13*mv22);
      Vector col2 = new Vector(mv23*mv31-mv21*mv33, mv11*mv33-mv13*mv31, mv13*mv21-mv11*mv23);
      Vector col3 = new Vector(mv21*mv32-mv22*mv31, mv12*mv31-mv11*mv32, mv11*mv22-mv12*mv21);
      Vector col4 = new Vector(0.0, 0.0, 0.0, 1.0);
      Matrix MVInvTranspose = new Matrix(col1, col2, col3, col4);


      // Walk this model's list of triangles.
      for (Triangle t : model.triangleList)
      {
         // Get all of the values that we need.
         double s = t.mtrl.s; //shininess coefficient
         double kar = t.mtrl.ambient[0];  // red
         double kdr = t.mtrl.diffuse[0];
         double ksr = t.mtrl.specular[0];
         double car = model.light.ambient[0];
         double cdr = model.light.diffuse[0];
         double csr = model.light.specular[0];
         double kag = t.mtrl.ambient[1];  // green
         double kdg = t.mtrl.diffuse[1];
         double ksg = t.mtrl.specular[1];
         double cag = model.light.ambient[1];
         double cdg = model.light.diffuse[1];
         double csg = model.light.specular[1];
         double kab = t.mtrl.ambient[2];  // blue
         double kdb = t.mtrl.diffuse[2];
         double ksb = t.mtrl.specular[2];
         double cab = model.light.ambient[2];
         double cdb = model.light.diffuse[2];
         double csb = model.light.specular[2];


         // Compute a color value at each of the three vertices of this triangle.
         for (int i = 0; i < 3; i++)
         {
            // To do vector algebra, we need the vertex as a vector.
            Vector p = new Vector(t.v[i].x, t.v[i].y, t.v[i].z);
            // We need the normal vector at this vertex.
            // Here is this vertex's normal vector in model(!) coordinates.
            Vector N = t.v[i].n;
            // We need to convert this normal vector into view coordinates.
            N = MVInvTranspose.times( N );
            N = N.normalize();
            // Put this normal vector into the vertex (for debugging purposes).
            t.v[i].n = N;
            // We need the unit light vector, L, from our vertex to the light source.
            // Here is the light source as a vertex in world coordinates.
            Vertex S = new Vertex(model.light.pos);
            // Convert light source from world coordinates to view coordinates.
            Matrix viewMatrix = model.viewMatrix;
            S = viewMatrix.times(S);
            // Convert the vertex to a vector.
            Vector S2 = new Vector(S.x, S.y, S.z);
            // Compute L = S2 - p, the vector from our vertex to the light source.
            Vector L = S2.plus( p.times(-1.0) );
            // normalize L
            L = L.normalize();
            // We need the reflection vector R = 2(L dot N)*N - L.
            Vector R = N.times( 2*L.dotProduct(N) ).plus( L.times(-1.0) );
            // We need the unit view vector, V, that points from the vertex to the origin.
            Vector V = p.normalize().times(-1.0);

            // Compute the color at this vertex using its normal, light, reflection, and view vectors,
            //    c = k_a * c_a + sum_{all lights}( k_d*c_d*(L dot N) + k_s*c_s*(R dot V)^s ) )

            // pre compute some numbers
            double diffuseAngle = L.dotProduct(N);
            diffuseAngle = (diffuseAngle < 0) ? 0.0 : diffuseAngle;
            double shininess = R.dotProduct(V);
            shininess = (shininess < 0) ? 0.0 : Math.pow(shininess, s);

            // do red
            double r = kar*car + kdr*cdr*diffuseAngle + ksr*csr*shininess;
            r = (r <= 1.0) ? r : 1.0;  // clamp the color value at 1.0
            t.v[i].r = r;

            // do green
            double g = kag*cag + kdg*cdg*diffuseAngle + ksg*csg*shininess;
            g = (g <= 1.0) ? g : 1.0;  // clamp the color value at 1.0
            t.v[i].g = g;

            // do blue
            double b = kab*cab + kdb*cdb*diffuseAngle + ksb*csb*shininess;
            b = (b <= 1.0) ? b : 1.0;  // clamp the color value at 1.0
            t.v[i].b = b;
         }
      }
   }
}