/**
   This is the "rendering pipeline".

   This program takes as its input a Scene data structure
   and it produces, as its output, a filled in FrameBuffer
   data structure. The final FrameBuffer holds the rendered
   image of the scene represented by the Scene object.
*/

public class SceneRender
{
   public static void renderScene(Scene scene, FrameBuffer fb)
   {
      Scene scene2 = new Scene(scene);  // Duplicate the Scene object,
                                        // so that we can modify the copy.
   /* NOTE: We could make it the client's responsibility
            to copy the Scene. In the case of OpenGL,
            the client "duplicates" the scene when it
            passes all of the scene information into the
            pipeline, and so all of the information gets
            copied into the graphics card's memory.
   */
      printSceneMatrices(scene2);

      for (Model model : scene2.modelList)
      {
         // Copy some information from the Scene to this Model.
         model.addViewMatrix(scene2.viewMatrix);
         model.addProjMatrix(scene2.projMatrix);
         model.addLight(scene2.light);

         printModelMatrices(model);
         printVertices(model, 1, 0);

         // apply the model transformation
         P1_Model2World.P1_Model2World(model);

         printVertices(model, 2, 1);

         // apply the view transformation
         P2_World2View.P2_World2View(model);

         printVertices(model, 4, 2);

         // delete back facing triangles
         P3_BackFaceCulling.P3_BackFaceCulling(model);

         printVertices(model, 8, 3);

         // compute vertex shading
         P4_Lighting.P4_Lighting(model);
         //P4_LightingSimple_ver1.P4_Lighting(model);
         //P4_LightingSimple_ver2.P4_Lighting(model);

         printVertices(model, 16, 4);

         // apply the projection transformation
         P5_View2Clip.P5_View2Clip(model);

         printVertices(model, 32, 5);

         // clip triangles to the standard volume
         P6_Clipping.P6_Clipping(model);
         //P6_ClippingSimple.P6_Clipping(model);

         printVertices(model, 64, 6);

         // transform to normalized device coordinates
         P7_Clip2NDC.P7_Clip2NDC(model);

         printVertices(model, 128, 7);

         // transform normalized device coordinates to viewport coordinates
         P8_NDC2VP.P8_NDC2VP(model, fb.getWidthVP(), fb.getHeightVP());

         printVertices(model, 256, 8);

         // rasterize each triangle into fragments (why not rasterize in NDC?)
         P9_Rasterize.P9_Rasterize(model);

         printVertices(model, 512, 9);

         // shade fragments of each triangle
         P10_FragmentShaderTexture.P10_FragmentShaderTexture(model);

         printVertices(model, 1024, 10);

         // use the fragments to fill in the viewport
         P11_Fragments2FB.P11_Fragments2FB(model, fb);
      }
   }//renderScene()


   /**
      Used for debugging.
   */
   public static void printSceneMatrices(Scene scene)
   {
      // Display the matrices stored in the Scene object.
      int printMatrices = 0;
      String matrixString = System.getenv("PrintMatrices");
      if (matrixString != null)
      {
         printMatrices = Integer.parseInt(matrixString); // 1 => print view matrix
      }                                                  // 2 => print projection matrix

      if ((printMatrices & 1) == 1)
      {
         System.err.println( scene.viewMatrix.toString() );
      }
      if ((printMatrices & 2) == 2)
      {
         System.err.println( scene.projMatrix.toString() );
      }
   }//printSceneMatrices()


   public static void printModelMatrices(Model model)
   {
      // Display the matrices stored in the Model object.
      int printMatrices = 0;
      String matrixString = System.getenv("PrintMatrices");
      if (matrixString != null)
      {
         printMatrices = Integer.parseInt(matrixString); // 4 => print model matrix
      }

      if ((printMatrices & 4) == 4)
      {
         System.err.println( model.modelMatrix.toString() );
      }
   }//printModelMatrices()


   /**
      Used for debugging.
   */
   public static void printVertices(Model model, int stageFlag, int stageNo)
   {
      int debugStages = 0;
      String debugString = null;
      if ( (debugString = System.getenv("PipeLine0")) != null ) debugStages += 1;
      if ( (debugString = System.getenv("PipeLine1")) != null ) debugStages += 2;
      if ( (debugString = System.getenv("PipeLine2")) != null ) debugStages += 4;
      if ( (debugString = System.getenv("PipeLine3")) != null ) debugStages += 8;
      if ( (debugString = System.getenv("PipeLine4")) != null ) debugStages += 16;
      if ( (debugString = System.getenv("PipeLine5")) != null ) debugStages += 32;
      if ( (debugString = System.getenv("PipeLine6")) != null ) debugStages += 64;
      if ( (debugString = System.getenv("PipeLine7")) != null ) debugStages += 128;
      if ( (debugString = System.getenv("PipeLine8")) != null ) debugStages += 256;
      if ( (debugString = System.getenv("PipeLine9")) != null ) debugStages += 512;
      if ( (debugString = System.getenv("PipeLine10")) != null ) debugStages += 1024;
      if ( (debugString = System.getenv("PipeLine11")) != null ) debugStages += 2048;
      //if (debugStages == 0) debugStages = 128; // default to NDC

      int printFields = 1;  // default to printing out vertex data
      String printString = System.getenv("PrintLevel");
      if (printString != null)
      {
         printFields = Integer.parseInt(printString); // 1 => print vertex data
      }                                               // 2 => print color data
                                                      // 4 => print texture coordinates
                                                      // 8 => print normal vector
      if ((stageFlag & debugStages) != 0)
      {
         System.err.printf("================= Stage P%d begin ===========\n", stageNo);
         System.err.print( model.toString(printFields) );
         System.err.printf("================= Stage P%d end =============\n", stageNo);
      }
   }//printVertices()
}